/* readstream.c -- read data from xyplot(1) format file
 * Copyright (C) 1994 David Deaven
 *
 * This file is part of xyplot.
 *
 * xyplot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * xyplot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with xyplot; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: readstream.c,v 1.5 1996/03/16 00:15:03 deaven Exp $
 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#include "dmd/memory.h"
#include "xyplot.h"

#include "site.h"

int breakflag = 0; /* the "break" token sets this */

static int xcol_read, ycol_read; /* last columns returned in multi-column read */
static int point,line;          /* current point, current line in a file */
static int comment_mode;        /* comment is being scanned */
static FILE *fin;		/* input stream */
static double *data = NULL;	/* data buffer */

/* Get the next token from the currently open stream
 * and return it, or NULL when no more tokens can be scanned.
 */
static char *buffer = NULL;
static int bsize = 0;
#define CHUNK 64;
static char *get_token(void)
{
  int nest = 0, i = 0, ic;
  while((ic = getc(fin)) != EOF)
  {
    if(i>=bsize) {
	bsize += CHUNK;
	buffer = myrealloc(buffer, bsize * sizeof(char));
    }
    switch(ic) {
      case '\n':
	++line;
	if(nest!=0) {
	  buffer[i]='\0';
	  fprintf(stderr,"WARNING: unbalanced ()'s -- ignoring `%s'\n", buffer);
	  return NULL;
	}
      case ' ':
      case '\t':
      case ',':
	if(i==0) break;
	if(nest==0) {buffer[i]='\0'; return buffer;}
	else {buffer[i++]=ic; break;}
      case '(': ++nest; buffer[i++]=ic; break;
      case ')': --nest; buffer[i++]=ic; break;
      default: buffer[i++]=ic; break;
    }
  } 
  if(i>=bsize) {
      bsize += CHUNK;
      buffer = myrealloc(buffer, bsize * sizeof(char));
  }
  buffer[i]='\0';
  return (i==0 ? NULL : buffer);
}

static void fixcol(XYFileInfo *fi, int *c)
{
  if (*c == XY_LAST_COLUMN) *c = fi->ncol;
  if (*c == XY_NO_COLUMN) *c = 0;
}

/* Open a data file and get number of columns.
 * The name field of the supplied structure must be filled in,
 * as well as the desired columns to read from.
 * Return non-zero error code, or zero for normal operation.
 */
int XYStartStream(XYFileInfo *fi)
{
  int n;
  char *tok;
  point=line=0;
  if((fin=fopen(fi->name,"r"))==NULL) {
    fprintf(stderr,"XYStartStream: can't open file `%s'\n",fi->name);
    return 1;
  }

  if((tok = get_token()) == NULL) {
    fprintf(stderr,"can't read number of columns in file `%s' from token `%s'\n",
      fi->name, tok);
    return 1;
  }
  n=0; XYintscan(tok, &n);
  if(n<=0) {
    fprintf(stderr,"bad number of columns in file `%s' (token `%s')\n",
      fi->name, tok);
    return 1;
  }
  fi->ncol=n;

  data = mymalloc(n * sizeof(double));

  /* Adjust column specs as neccesary
   */
  fixcol(fi, &(fi->xcol));
  fixcol(fi, &(fi->ycol));
  fixcol(fi, &(fi->xerr));
  fixcol(fi, &(fi->yerr));
  fixcol(fi, &(fi->xcolmax));
  fixcol(fi, &(fi->ycolmax));

  /* check column specifications
   */
  if((fi->xcol<1) || (fi->ycol<1)) {
    fprintf(stderr,"bad column specifications for file `%s'\n",fi->name);
    return 1;
  }
  if((fi->xcol>n) || (fi->ycol>n)) {
    fprintf(stderr,
      "column specification out of range for file `%s' (%d columns available)\n",
      fi->name,n);
    return 1;
  } 

  xcol_read = fi->xcolmax; /* mark as read so next time starts */
  ycol_read = fi->ycolmax;

  return 0;
}

/* Execute a special token. 
 */
static void do_token(XYFileInfo *fi, char *s)
{
  int i,j,n;
  if(comment_mode) {  /* just watch for a token ending in */
    if((i=strlen(s))<2)return;
    if(s[i-1]=='/' && s[i-2]=='*')comment_mode=0;
    return;
  }
  switch(s[0]) {
    case '\0': return;
    case '#': /* data column label */
      i=0; while((s[i]!='=') && (s[i]!='\0'))++i;
      if(s[i]=='\0')break; s[i]='\0';
      n=0; XYintscan(s+1,&n); if(n==0)break;
      s+=i+1;
      if(s[0]!='(')break;
      j=strlen(s); if(s[j-1]!=')')break; s[j-1]='\0';
      if(n==fi->xcol)fi->xlabel=mystrdup(s+1);
      if(n==fi->ycol)fi->ylabel=mystrdup(s+1);
      return;
    case '/': /* PostScript variable definition, or start of comment */
      if(s[1]=='*') { comment_mode=1; return; }
      i=0; while((s[i]!='=') && (s[i]!='\0'))++i;
      if(s[i]=='\0') { /* not an assignment, just a command */
        printf("%s\n",s+1);
      } else {
        s[i]='\0';
        if(strcmp(s,"/lab")==0) { /* special variable, intercepted */
	  s+=i+2; i=strlen(s)-1; s[i]='\0';
	  fi->label=mystrdup(s);
	} else printf("%s %s def\n",s,s+i+1); 
      }
      return;
    case 'b':
      if(strcmp(s, "break")) break;
      breakflag = 1;
      return;
    default: break;
  }
  fprintf(stderr,"do_token(): bad token `%s'\n",s);
  exit(1);
}

/* Read all of the next data columns from the input stream. 
 * Return 0 or EOF (file is closed here).
 */
int XYReadData(XYFileInfo *fi, double *d)
{
  double f;
  int col = 0;
  char *tok;
  static char *okdigit = "0123456789.-+n";
  while(col<fi->ncol) {
    if((tok = get_token()) == NULL || tok[0] == '!') {
      fclose(fin);
      myfree(data);
      data = NULL;
      return EOF;
    }
#if 0
    if((!isdigit(tok[0]) && tok[0] != '-' && tok[0] != '+'
         && tok[0] != 'n') || comment_mode)
#else
    if(strchr(okdigit, *tok) == NULL)
#endif
      do_token(fi, tok);
    else {
      f=0.0;
      if(XYfloatscan(tok, &f)) {
	if(strcmp(tok, "null")) /* just a bad entry */
	  fprintf(stderr,
  "WARNING: assuming 0 for bad data `%s' near point %d (line %d) in file `%s'\n",
	    tok, point,line,fi->name);
	else f=XY_NULL; /* a null entry -- skip it */
      }
      if(f!=XY_NULL) {
	if(fabs(f)<PS_SMALL)f=0.0;
	if(fabs(f)>PS_BIG) {
	  f=copysign(PS_BIG,f);
	  fprintf(stderr,"WARNING: changing token `%s' to %.6g\n", tok, f);
	}
      }
      d[col++]=f;
    }
  }
  return 0;
}

/* Read the next data from the input stream. 
 * Return 0 or EOF.
 * If multiple columns are selected as equal, return them one at
 * a time.  ?col_read always points at the next data to be read.
 */
int XYReadStream(XYFileInfo *fi,
  double *x, double *y, double *xerr, double *yerr)
{
  do {
    if(xcol_read >= fi->xcolmax &&
       ycol_read >= fi->ycolmax    ) /* read new data */
    {
      if(XYReadData(fi, data)==EOF) return EOF;
      xcol_read = fi->xcol;
      ycol_read = fi->ycol;
    }
    else /* advance the counters */
    {
      if(xcol_read < fi->xcolmax) ++xcol_read;
      else if(ycol_read < fi->ycolmax) ++ ycol_read;
    }
    *x = data[xcol_read - 1];
    *y = data[ycol_read - 1];
  } while( (*x)==XY_NULL || (*y)==XY_NULL);
  if(fi->xerr>0)*xerr=data[fi->xerr-1];
  else *xerr=0.0;
  if(fi->yerr>0)*yerr=data[fi->yerr-1];
  else *yerr=0.0;
  ++point;
  return 0;
}
