/* xyplot -- make a PostScript file to generate an xy plot, from some
 *           input files that contain some data.
 * Copyright (C) 1994 David Deaven
 *
 * This file is part of xyplot.
 *
 * xyplot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * xyplot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with xyplot; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: xyplot.c,v 1.4 1995/10/10 15:03:39 deaven Exp $
 */
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include "dmd/memory.h"
#include "xyplot.h"

#include "site.h"

static char *usage=
  "xyplot version 1.1, Copyright (C) 1994 David Deaven\n"
  "xyplot comes with ABSOLUTELY NO WARRANTY; for details see\n"
  "the GNU General Public License version 2 or greater.\n"
  "This is free software, and you are welcome to redistribute it\n"
  "under certain conditions (see the GNU license for details).\n\n"

  "usage: %s <options> file x y [file x y ...]\n"
  "options:\n"
  "\t-a               omit plot axes\n"
  "\t-c               toggle connect mode\n"
  "\t-d               output data only\n"
  "\t-f               omit plot frame\n"
  "\t-h               omit header file\n"
  "\t-L[x|y]          use logarithmic axes\n"
  "\t-l               omit plot legend\n"
  "\t-n #             set number of tick marks\n"
  "\t-o file          write output to file\n"
  "\t-u               enable UNIX plot(1) mode\n"
  "\t-R[x|y] min max  set range of x values to plot\n"
  "\t-s               omit showpage PS command\n"
  "\t-t               toggle timestamp mode\n"
  "\t-v               be verbose\n"
  "\t-[x|y] size      set plot size in mm\n"
  ;

static char *markername[]= {
  "square outline", "diamond outline", "circle outline", "triangle outline",
  "square sfill", "diamond sfill", "circle sfill", "triangle sfill",
  "star outline", "star sfill", "cross stroke",
  NULL
};

static char *lstylename[] = {
  "[] 0 setdash", "[ms] 0 setdash", "[ms 2 mul] 0 setdash",
  "[ms 2 div] 0 setdash", "[ms 2 mul ms 2 div dup dup] 0 setdash",
  "[ms 2 mul ms 2 div dup dup dup dup] 0 setdash",
  NULL
};

extern int breakflag; /* stream reader can set this on "break" token */

#define XYPLOT_HEADER "xyplot.ps"
char *xyplot_path;

/* default command-line alterable parameters */

int datamode=0;   /* write just a list of x,y data flag */
int header=1;     /* insert the header file */
int showpage=1;   /* include a "showpage" command at the end */
int ntick=5;      /* target number of tick marks to put on axis */
int conn=0;       /* use markers, not line segments */
int unixmode=0;   /* UNIX graph() compatibility mode */
int log_x=0;      /* X axis logarithmic */
int log_y=0;      /* Y axis logarithmic */
int frame=1;      /* put a frame on the plot */
int stampit=0;    /* time stamp */
int makeaxes=1;   /* make x and y axis instructions */
int legend=1;     /* make legend instructions */
double psizex=130; /* horizontal plot size in mm */
double psizey=150; /* vertical plot size in mm */
int verbose=0;    /* information you probably don't want */

/* misc global variables */

double xmin=PS_BIG, xmax=-PS_BIG, ymin=PS_BIG, ymax=-PS_BIG;

void main(int argc, char **argv)
{
  int arg,i,j;
  int marker,lstyle;
  double umin,ustp,umax;
  double vmin,vstp,vmax;
  char *timestamp,*xlabel,*ylabel,*label;
  int xlock=0,ylock=0;
  char *ofn = "-"; /* stdout */
  FILE *fout;
  XYFileInfo *FileInfo = NULL;
  int NFile = 0;

/* get system date */

  timestamp=GetTime();

/* Read command line and perform pre-processing */

  for(arg=1; arg<argc; ++arg) {
    if(argv[arg][0]=='-') {  /* this is an input option */
      switch(argv[arg][1]) {
        case 'n': XYintscan(&argv[arg][2],&ntick); break;
	case 'x': XYfloatscan(&argv[arg][2],&psizex); break;
	case 'y': XYfloatscan(&argv[arg][2],&psizey); break;
        case 'd': datamode=1;
		  unixmode=header=legend=makeaxes=0; break;
	case 'h': header=!header; break;
	case 's': showpage=!showpage; break;
	case 'a': makeaxes=!makeaxes; break;
	case 't': stampit=!stampit; break;
	case 'l': legend=!legend; break;
	case 'o': ofn = argv[++arg]; break;
	case 'c': conn=!conn; break;
	case 'u': unixmode=1; 
		  datamode=header=legend=makeaxes=0; break;
	case 'f': frame=!frame; break;
	case 'v': verbose=!verbose; break;
	case 'L': 
	  if(argv[arg][2]=='x') { log_x=!log_x; break; }
	  else if(argv[arg][2]=='y'){ log_y=!log_y; break; }
	case 'R':
	  if((arg+2)>=argc) {
	    fprintf(stderr,"need two arguments for %s\n",argv[arg]);
	    exit(1);
	  }
	  if(argv[arg][2]=='x') {
	    XYfloatscan(argv[++arg],&xmin);
	    XYfloatscan(argv[++arg],&xmax);
	    xlock=1;
	  } else if(argv[arg][2]=='y') {
	    XYfloatscan(argv[++arg],&ymin);
	    XYfloatscan(argv[++arg],&ymax);
	    ylock=1;
	  } else {
	    fprintf(stderr,"set range for x and y only (%s)\n",argv[arg]);
	    exit(1);
	  }
	  break;
        default:
	  fprintf(stderr,"unrecognizable option %s\n",argv[arg]);
	case 'V':
	  fprintf(stderr,usage,argv[0]);
	  exit(1);
      }
    } else { /* this is a filename */ 
      FileInfo = myrealloc(FileInfo, (NFile + 1) * sizeof(XYFileInfo));

      if(strcmp(argv[arg],"+")==0)
      {  /* use the last filename */
	if(NFile<1) {
	  fprintf(stderr,"I need a filename before you use +\n");
	  exit(1);
	}
	FileInfo[NFile].name=FileInfo[NFile-1].name;
      }
      else
        FileInfo[NFile].name=argv[arg];

      if((arg+2)>=argc) {
        fprintf(stderr,"specify two columns to plot for %s\n",argv[arg]);
	exit(1);
      }
      ScanCol(argv[++arg],
        &FileInfo[NFile].xcol,&FileInfo[NFile].xcolmax,&FileInfo[NFile].xerr);
      ScanCol(argv[++arg],
        &FileInfo[NFile].ycol,&FileInfo[NFile].ycolmax,&FileInfo[NFile].yerr);

      FileInfo[NFile].connect=conn;

      FileInfo[NFile].xlabel =
        DefaultLabel(FileInfo[NFile].xerr, FileInfo[NFile].xcol);
      FileInfo[NFile].ylabel =
        DefaultLabel(FileInfo[NFile].yerr, FileInfo[NFile].ycol);
      FileInfo[NFile].label="";
      if(XYScanRange(FileInfo + NFile)) exit(1);
      ++NFile;
    }
  }

  if(NFile < 1) {
  	fprintf(stderr, "%s: no files to plot\n", argv[0]);
  	exit(1);
  }

  fout = (strcmp(ofn, "-") == 0 ? stdout : fopen(ofn, "w"));
  if(fout == NULL) {
  	fprintf(stderr, "%s: cannot open %s\n", argv[0], ofn);
  	exit(1);
  }

  xyplot_path = getenv("XYPLOTPATH");
  if(xyplot_path == NULL)
    xyplot_path = DEFAULT_XYPLOTPATH;
  else
    xyplot_path = mystrdup(xyplot_path);

/* Determine plot range and axis labels */

  if(verbose) {
    for(i=0;i<NFile;++i) {
      fprintf(stderr,"File %s col %d (%f,%f) col %d (%f,%f)\n",
	FileInfo[i].name,
	FileInfo[i].xcol,FileInfo[i].xmin,FileInfo[i].xmax,
	FileInfo[i].ycol,FileInfo[i].ymin,FileInfo[i].ymax);
    }
  }
  if(xmin==PS_BIG)
    for(i=0;i<NFile;++i)
      if(FileInfo[i].xmin<xmin)xmin=FileInfo[i].xmin;
  if(xmax==-PS_BIG)
    for(i=0;i<NFile;++i)
      if(FileInfo[i].xmax>xmax)xmax=FileInfo[i].xmax;
  if(ymin==PS_BIG)
    for(i=0;i<NFile;++i)
      if(FileInfo[i].ymin<ymin)ymin=FileInfo[i].ymin;
  if(ymax==-PS_BIG)
    for(i=0;i<NFile;++i)
      if(FileInfo[i].ymax>ymax)ymax=FileInfo[i].ymax;
  if(verbose)fprintf(stderr,
    "plot range: (%f,%f),(%f,%f)\n",xmin,xmax,ymin,ymax);
  xlabel=FileInfo[0].xlabel;
  ylabel=FileInfo[0].ylabel;

  /* Construct a default label for the plot.  Try to list
   * the unique labels of all of the files, unless that would be very
   * long -- then just indicate that there are many files.
   */
  {
    int l = 0;
    int *use = mymalloc(NFile * sizeof(int));
    for(i=0; i<NFile; ++i)
    {
      for(j=0; j<i; ++j)
        if(strcmp(FileInfo[i].label, FileInfo[j].label) == 0) break;
      if(j<i)
        use[i] = 0;
      else
      {
	use[i] = 1;
	l += 2 + (FileInfo[i].label[0] == '\0' ? 
	  strlen(FileInfo[i].name) + 2 : strlen(FileInfo[i].label));
      }
    }
    if(l < 80)
    {
      label = mymalloc(l * sizeof(char));
      label[0] = '\0';
      for(i=0; i<NFile; ++i)
      {
	if(use[i])
	{
	  if(FileInfo[i].label[0] == '\0')
	  {
	    strcat(label, "\"");
	    strcat(label, FileInfo[i].name);
	    strcat(label, "\"");
	  }
	  else
	    strcat(label, FileInfo[i].label);
	  strcat(label, ", ");
	}
      }
      label[l - 2] = '\0';
    } 
    else
    {
      label = mymalloc((32 + strlen(FileInfo[0].name) +
	strlen(FileInfo[NFile-1].name)) * sizeof(char));
      sprintf(label, "%s, %s, ... (%d files)", 
	FileInfo[0].name, FileInfo[NFile - 1].name, NFile);
    }
    myfree(use);
  }

  if(verbose)fprintf(stderr, "plot label: %s\n",label);

/* Generate output */

  if(header) {
    fprintf(fout, "%%!PS-Adobe-\n"); /* minimally conforming PostScript */
    fprintf(fout, "%%%%DocumentFonts: Times-Roman\n");
    fprintf(fout, "%%%%Creator: xyplot\n");
    fprintf(fout, "%%%%CreationDate: %s\n",timestamp);
    fprintf(fout, "%%%%BoundingBox: 100 100 %.1f %.1f\n",
      (psizex + 51.0) * 2.835, (psizey + 51.0) * 2.835);
    fprintf(fout, "%%%%EndComments\n");
    if(finsert(fout, XYPLOT_HEADER)!=0) {
      fprintf(stderr,"cannot insert header file %s\n",XYPLOT_HEADER);
      exit(1);
    }
  }
  if(!datamode && !unixmode) {
    fprintf(fout, "xyplotDict begin xyinit\n");
    if(header)fprintf(fout, "50 mm 50 mm translate\n");
    fprintf(fout, "/psizex %.6g mm def /psizey %.6g mm def\n",psizex,psizey);
  }
  if(makeaxes) {
    fprintf(fout, "/lab {(%s) showcenter} def\n",label);
    fprintf(fout, "/xlab {(%s) showcenter} def\n",xlabel);
    fprintf(fout, "/ylab {(%s) showcenter} def\n",ylabel);
    i=ticker(xmin,xmax,ntick,xlock,&umin,&ustp,&umax);
    j=ticker(ymin,ymax,ntick,ylock,&vmin,&vstp,&vmax);
    fprintf(fout, "/xmin %.6g def /xmax %.6g def /axisx xmin def\n",umin,umax);
    fprintf(fout, "/ymin %.6g def /ymax %.6g def /axisy ymin def\n",vmin,vmax);
    if(frame)fprintf(fout, "makeframe\n");
    MakeTicks(fout, i,umin,ustp,umax,'x');
    MakeTicks(fout, j,vmin,vstp,vmax,'y');
    fprintf(fout, "makelabel\n");
  } else if(!datamode && !unixmode) {
    fprintf(fout, "/xmin %.6g def /xmax %.6g def /axisx xmin def\n",xmin,xmax);
    fprintf(fout, "/ymin %.6g def /ymax %.6g def /axisy ymin def\n",ymin,ymax);
    if(frame)fprintf(fout, "makeframe\n");
  }
  if(legend)fprintf(fout, "(_new) legend\n");

  marker=lstyle=0;
  if(unixmode)setUnix(fout, xmin,ymin,xmax,ymax); /* clear screen, set range */
  for(i=0;i<NFile;++i) {
    if(XYStartStream(FileInfo + i)) exit(1);
    if(datamode)DumpData(fout, FileInfo + i);
    else if(unixmode)DumpUnix(fout, FileInfo + i);
    else {
      fprintf(fout, "%% file %s: column %d vs %d. (%s vs %s)\n",
	FileInfo[i].name,FileInfo[i].ycol,FileInfo[i].xcol,
	FileInfo[i].ylabel,FileInfo[i].xlabel);
      if(FileInfo[i].connect) {
	fprintf(fout, "/dataline {stdline %s} def /marker /dataline def\n",
	  lstylename[lstyle++]);
	if(lstylename[lstyle] == NULL)lstyle=0;
	if(legend)fprintf(fout, "(%s vs. %s from %s) legend\n",
	  FileInfo[i].ylabel,FileInfo[i].xlabel,FileInfo[i].name);
	fprintf(fout, "beginclip startline\n");
	DumpPlot(fout, FileInfo + i,'c');
        fprintf(fout, "stroke endclip\n");
      } else {
	fprintf(fout, "/marker {startmarker %s} def\n",markername[marker++]);
	if(markername[marker] == NULL) marker=0;
	if(legend)fprintf(fout, "(%s vs. %s from %s) legend\n",
	  FileInfo[i].ylabel,FileInfo[i].xlabel,FileInfo[i].name);
	DumpPlot(fout, FileInfo + i,'p');
      }
    }
  }
  if(!datamode && !unixmode) {
    if(legend)fprintf(fout, "makelegend\n");
    if(showpage)fprintf(fout, "showpage\n");
    fprintf(fout, "end\n");
  }

  if(fout != stdout) fclose(fout);
}
